import asyncio
import errno
import functools
import grp
import os
import pwd
import shutil
import signal
import sys
import tempfile

from traitlets import Integer, List, Unicode

from ..traitlets import Type
from .base import ClusterConfig
from .db_base import DBBackendBase

__all__ = ("LocalClusterConfig", "LocalBackend", "UnsafeLocalBackend")


class LocalClusterConfig(ClusterConfig):
    """Dask cluster configuration options when running as local processes"""

    pass


def _signal(pid, sig):
    """Send given signal to a pid.

    Returns True if the process still exists, False otherwise."""
    try:
        os.kill(pid, sig)
    except OSError as e:
        if e.errno == errno.ESRCH:
            return False
        raise
    return True


def is_running(pid):
    return _signal(pid, 0)


async def wait_is_shutdown(pid, timeout=10):
    """Wait for a pid to shutdown, using exponential backoff"""
    pause = 0.1
    while timeout >= 0:
        if not _signal(pid, 0):
            return True
        await asyncio.sleep(pause)
        timeout -= pause
        pause *= 2
    return False


@functools.lru_cache
def getpwnam(username):
    return pwd.getpwnam(username)


class LocalBackend(DBBackendBase):
    """A cluster backend that launches local processes.

    Requires super-user permissions in order to run processes for the
    requesting username.
    """

    cluster_config_class = Type(
        "dask_gateway_server.backends.local.LocalClusterConfig",
        klass="dask_gateway_server.backends.base.ClusterConfig",
        help="The cluster config class to use",
        config=True,
    )

    sigint_timeout = Integer(
        10,
        help="""
        Seconds to wait for process to stop after SIGINT.

        If the process has not stopped after this time, a SIGTERM is sent.
        """,
        config=True,
    )

    sigterm_timeout = Integer(
        5,
        help="""
        Seconds to wait for process to stop after SIGTERM.

        If the process has not stopped after this time, a SIGKILL is sent.
        """,
        config=True,
    )

    sigkill_timeout = Integer(
        5,
        help="""
        Seconds to wait for process to stop after SIGKILL.

        If the process has not stopped after this time, a warning is logged and
        the process is deemed a zombie process.
        """,
        config=True,
    )

    clusters_directory = Unicode(
        help="""
        The base directory for cluster working directories.

        A subdirectory will be created for each new cluster which will serve as
        the working directory for that cluster. On cluster shutdown the
        subdirectory will be removed.

        If not specified, a temporary directory will be used for each cluster.
        """,
        config=True,
    )

    inherited_environment = List(
        [
            "PATH",
            "PYTHONPATH",
            "CONDA_ROOT",
            "CONDA_DEFAULT_ENV",
            "VIRTUAL_ENV",
            "LANG",
            "LC_ALL",
        ],
        help="""
        Whitelist of environment variables for the scheduler and worker
        processes to inherit from the Dask-Gateway process.
        """,
        config=True,
    )

    default_host = "127.0.0.1"

    def set_file_permissions(self, paths, username):
        pwnam = getpwnam(username)
        for p in paths:
            os.chown(p, pwnam.pw_uid, pwnam.pw_gid)

    def make_preexec_fn(self, cluster):  # pragma: nocover
        # Borrowed and modified from jupyterhub/spawner.py
        pwnam = getpwnam(cluster.username)
        uid = pwnam.pw_uid
        gid = pwnam.pw_gid
        groups = [g.gr_gid for g in grp.getgrall() if cluster.username in g.gr_mem]
        workdir = cluster.state["workdir"]

        def preexec():
            os.setgid(gid)
            try:
                os.setgroups(groups)
            except Exception as e:
                print("Failed to set groups %s" % e, file=sys.stderr)
            os.setuid(uid)
            os.chdir(workdir)

        return preexec

    def setup_working_directory(self, cluster):  # pragma: nocover
        if self.clusters_directory:
            workdir = os.path.join(self.clusters_directory, cluster.name)
        else:
            workdir = tempfile.mkdtemp(prefix="dask", suffix=cluster.name)
        certsdir = self.get_certs_directory(workdir)
        logsdir = self.get_logs_directory(workdir)

        paths = [workdir, certsdir, logsdir]
        for path in paths:
            os.makedirs(path, 0o700, exist_ok=True)

        cert_path, key_path = self._get_tls_paths(workdir)
        flags = os.O_WRONLY | os.O_CREAT | os.O_EXCL
        for path, data in [(cert_path, cluster.tls_cert), (key_path, cluster.tls_key)]:
            with os.fdopen(os.open(path, flags, 0o600), "wb") as fil:
                fil.write(data)
                paths.extend(path)

        self.set_file_permissions(paths, cluster.username)

        self.log.debug(
            "Working directory %s for cluster %s created", workdir, cluster.name
        )
        return workdir

    def cleanup_working_directory(self, workdir):
        if os.path.exists(workdir):
            try:
                shutil.rmtree(workdir)
                self.log.debug("Working directory %s removed", workdir)
            except Exception:  # pragma: nocover
                self.log.warn("Failed to remove working directory %r", workdir)

    def get_certs_directory(self, workdir):
        return os.path.join(workdir, ".certs")

    def get_logs_directory(self, workdir):
        return os.path.join(workdir, "logs")

    def _get_tls_paths(self, workdir):
        certsdir = self.get_certs_directory(workdir)
        cert_path = os.path.join(certsdir, "dask.crt")
        key_path = os.path.join(certsdir, "dask.pem")
        return cert_path, key_path

    def get_tls_paths(self, cluster):
        return self._get_tls_paths(cluster.state["workdir"])

    def get_env(self, cluster):
        env = super().get_env(cluster)
        for key in self.inherited_environment:
            if key in os.environ:
                env[key] = os.environ[key]
        env["USER"] = cluster.username
        return env

    async def start_process(self, cluster, cmd, env, name):
        workdir = cluster.state["workdir"]
        logsdir = self.get_logs_directory(workdir)
        log_path = os.path.join(logsdir, name + ".log")
        flags = os.O_WRONLY | os.O_CREAT | os.O_EXCL
        fd = None
        try:
            fd = os.open(log_path, flags, 0o755)
            proc = await asyncio.create_subprocess_exec(
                *cmd,
                preexec_fn=self.make_preexec_fn(cluster),
                start_new_session=True,
                env=env,
                stdout=fd,
                stderr=asyncio.subprocess.STDOUT,
            )
        finally:
            if fd is not None:
                os.close(fd)
        return proc.pid

    async def stop_process(self, pid):
        methods = [
            ("SIGINT", signal.SIGINT, self.sigint_timeout),
            ("SIGTERM", signal.SIGTERM, self.sigterm_timeout),
            ("SIGKILL", signal.SIGKILL, self.sigkill_timeout),
        ]

        for msg, sig, timeout in methods:
            self.log.debug("Sending %s to process %d", msg, pid)
            _signal(pid, sig)
            if await wait_is_shutdown(pid, timeout):
                return

        if is_running(pid):
            # all attempts failed, zombie process
            self.log.warn("Failed to stop process %d", pid)

    async def do_start_cluster(self, cluster):
        workdir = self.setup_working_directory(cluster)
        yield {"workdir": workdir}

        pid = await self.start_process(
            cluster,
            self.get_scheduler_command(cluster),
            self.get_scheduler_env(cluster),
            "scheduler",
        )
        yield {"workdir": workdir, "pid": pid}

    async def do_stop_cluster(self, cluster):
        pid = cluster.state.get("pid")
        if pid is not None:
            await self.stop_process(pid)

        workdir = cluster.state.get("workdir")
        if workdir is not None:
            self.cleanup_working_directory(workdir)

    def _check_status(self, o):
        pid = o.state.get("pid")
        return pid is not None and is_running(pid)

    async def do_check_clusters(self, clusters):
        return [self._check_status(c) for c in clusters]

    async def do_start_worker(self, worker):
        cmd = self.get_worker_command(worker.cluster, worker.name)
        env = self.get_worker_env(worker.cluster)
        pid = await self.start_process(
            worker.cluster, cmd, env, "worker-%s" % worker.name
        )
        yield {"pid": pid}

    async def do_stop_worker(self, worker):
        pid = worker.state.get("pid")
        if pid is not None:
            await self.stop_process(pid)

    async def do_check_workers(self, workers):
        return [self._check_status(w) for w in workers]


class UnsafeLocalBackend(LocalBackend):
    """A version of LocalBackend that doesn't set permissions.

    FOR TESTING ONLY! This provides no user separations - clusters run with the
    same level of permission as the gateway.
    """

    def make_preexec_fn(self, cluster):
        workdir = cluster.state["workdir"]

        def preexec():  # pragma: nocover
            os.chdir(workdir)

        return preexec

    def set_file_permissions(self, paths, username):
        pass



import math
import shutil
import socket

from traitlets import Bool, Unicode, default

from ...traitlets import Type
from .base import JobQueueBackend, JobQueueClusterConfig

__all__ = ("PBSBackend", "PBSClusterConfig")


def qsub_format_memory(n):
    """Format memory in bytes for use in qsub resources."""
    if n >= 10 * (1024**3):
        return "%dGB" % math.ceil(n / (1024**3))
    if n >= 10 * (1024**2):
        return "%dMB" % math.ceil(n / (1024**2))
    if n >= 10 * 1024:
        return "%dkB" % math.ceil(n / 1024)
    return "%dB" % n


def format_resource_list(template, cores, memory):
    return template.format(cores=cores, memory=qsub_format_memory(memory))


class PBSClusterConfig(JobQueueClusterConfig):
    """Dask cluster configuration options when running on PBS"""

    queue = Unicode("", help="The queue to submit jobs to.", config=True)

    account = Unicode(
        "", help="Accounting string associated with each job.", config=True
    )

    project = Unicode("", help="Project associated with each job.", config=True)

    worker_resource_list = Unicode(
        "select=1:ncpus={cores}:mem={memory}",
        help="""
        The resource list to use for the workers.

        This is a template, and receives the following fields:

        - cores
        - memory
        """,
        config=True,
    )

    scheduler_resource_list = Unicode(
        "select=1:ncpus={cores}:mem={memory}",
        help="""
        The resource list to use for the scheduler.

        This is a template, and receives the following fields:

        - cores
        - memory
        """,
        config=True,
    )

    use_stagein = Bool(
        True,
        help="""
        If true, the staging directory created above will be copied into the
        job working directories at runtime using the ``-Wstagein`` directive.

        If the staging directory is on a networked filesystem, you can set this
        to False and rely on the networked filesystem for access.
        """,
        config=True,
    )


class PBSBackend(JobQueueBackend):
    """A backend for deploying Dask on a PBS cluster."""

    cluster_config_class = Type(
        "dask_gateway_server.backends.jobqueue.pbs.PBSClusterConfig",
        klass="dask_gateway_server.backends.base.ClusterConfig",
        help="The cluster config class to use",
        config=True,
    )

    gateway_hostname = Unicode(
        help="""
        The hostname of the node running the gateway. Used for referencing the
        local host in PBS directives.
        """,
        config=True,
    )

    @default("gateway_hostname")
    def _default_gateway_hostname(self):
        return socket.gethostname()

    @default("submit_command")
    def _default_submit_command(self):
        return shutil.which("qsub") or "qsub"

    @default("cancel_command")
    def _default_cancel_command(self):
        return shutil.which("qdel") or "qdel"

    @default("status_command")
    def _default_status_command(self):
        return shutil.which("qstat") or "qstat"

    def get_tls_paths(self, cluster):
        if cluster.config.use_stagein:
            return "dask.crt", "dask.pem"
        return super().get_tls_paths(cluster)

    def get_submit_cmd_env_stdin(self, cluster, worker=None):
        cmd = [self.submit_command]
        cmd.extend(["-N", "dask-gateway"])
        if cluster.config.queue:
            cmd.extend(["-q", cluster.config.queue])
        if cluster.config.account:
            cmd.extend(["-A", cluster.config.account])
        if cluster.config.project:
            cmd.extend(["-P", cluster.config.project])
        cmd.extend(["-j", "eo", "-R", "eo", "-Wsandbox=PRIVATE"])

        if cluster.config.use_stagein:
            staging_dir = self.get_staging_directory(cluster)
            cmd.append(f"-Wstagein=.@{self.gateway_hostname}:{staging_dir}/*")

        if worker:
            resources = format_resource_list(
                cluster.config.worker_resource_list,
                cluster.config.worker_cores,
                cluster.config.worker_memory,
            )
            script = "\n".join(
                [
                    cluster.config.worker_setup,
                    " ".join(self.get_worker_command(cluster, worker.name)),
                ]
            )
            env = self.get_worker_env(cluster)
        else:
            resources = format_resource_list(
                cluster.config.scheduler_resource_list,
                cluster.config.scheduler_cores,
                cluster.config.scheduler_memory,
            )
            script = "\n".join(
                [
                    cluster.config.scheduler_setup,
                    " ".join(self.get_scheduler_command(cluster)),
                ]
            )
            env = self.get_scheduler_env(cluster)

        cmd.extend(["-v", ",".join(sorted(env))])
        cmd.extend(["-l", resources])
        cmd.extend(["--", "/bin/sh", "-c", script])

        return cmd, env, None

    def get_stop_cmd_env(self, job_id):
        return [self.cancel_command, job_id], {}

    def get_status_cmd_env(self, job_ids):
        out = [self.status_command, "-x"]
        out.extend(job_ids)
        return out, {}

    def parse_job_id(self, stdout):
        return stdout.strip()

    def parse_job_states(self, stdout):
        lines = stdout.splitlines()[2:]

        states = {}

        for l in lines:
            parts = l.split()
            job_id = parts[0]
            status = parts[4]
            states[job_id] = status in ("R", "Q", "H")

        return states



import base64
import os
import time
import uuid
from urllib.parse import quote

import aiohttp
from aiohttp import web
from traitlets import Instance, Integer, Unicode, default
from traitlets.config import LoggingConfigurable

from .models import User

__all__ = (
    "Authenticator",
    "SimpleAuthenticator",
    "KerberosAuthenticator",
    "JupyterHubAuthenticator",
)


class UserCache:
    def __init__(self, max_age):
        self.max_age = max_age
        self.name_to_cookie = {}
        self.cookie_to_user = {}

    def get(self, cookie):
        user = self.cookie_to_user.get(cookie)
        if user is None:
            return None
        timestamp, user = self.cookie_to_user[cookie]
        if timestamp + self.max_age < time.monotonic():
            del self.cookie_to_user[cookie]
            del self.name_to_cookie[user.name]
            return None
        return user

    def put(self, user):
        cookie = self.name_to_cookie.get(user.name)
        if cookie is None:
            cookie = uuid.uuid4().hex
            now = time.monotonic()
            self.name_to_cookie[user.name] = cookie
            self.cookie_to_user[cookie] = (now, user)
        return cookie


class Authenticator(LoggingConfigurable):
    """Base class for authenticators.

    An authenticator manages authenticating user API requests.

    Subclasses must define ``authenticate``, and may optionally also define
    ``setup``, ``cleanup`` and ``pre_response``.
    """

    cookie_name = Unicode(
        help="The cookie name to use for caching authentication information.",
        config=True,
    )

    @default("cookie_name")
    def _default_cookie_name(self):
        return "dask-gateway-%s" % uuid.uuid4().hex

    cache_max_age = Integer(
        300,
        help="""The maximum time in seconds to cache authentication information.

        Helps reduce load on the backing authentication service by caching
        responses between requests. After this time the user will need to be
        reauthenticated before making additional requests (note this is usually
        transparent to the user).
        """,
        config=True,
    )

    cache = Instance(UserCache)

    @default("cache")
    def _default_cache(self):
        return UserCache(max_age=self.cache_max_age)

    async def authenticate_and_handle(self, request, handler):
        # Try to authenticate with the cookie first
        cookie = request.cookies.get(self.cookie_name)
        if cookie is not None:
            user = self.cache.get(cookie)
            if user is not None:
                request["user"] = user
                return await handler(request)

        # Otherwise go through full authentication process
        user = await self.authenticate(request)
        if type(user) is tuple:
            user, context = user
        else:
            context = None

        request["user"] = user
        response = await handler(request)

        await self.pre_response(request, response, context)
        cookie = self.cache.put(user)
        response.set_cookie(self.cookie_name, cookie, max_age=self.cache_max_age * 2)

        return response

    async def setup(self, app):
        """Called when the server is starting up.

        Do any initialization here.

        Parameters
        ----------
        app : aiohttp.web.Application
            The aiohttp application. Can be used to add additional routes if
            needed.
        """
        pass

    async def cleanup(self):
        """Called when the server is shutting down.

        Do any cleanup here."""
        pass

    async def authenticate(self, request):
        """Perform the authentication process.

        Parameters
        ----------
        request : aiohttp.web.Request
            The current request.

        Returns
        -------
        user : User
            The authenticated user.
        context : object, optional
            If necessary, may optionally return an opaque object storing
            additional context needed to complete the authentication process.
            This will be passed to ``pre_response``.
        """
        raise NotImplementedError

    async def pre_response(self, request, response, context=None):
        """Called before returning a response.

        Allows modifying the outgoing response in-place to add additional
        headers, etc...

        Note that this is only called if ``authenticate`` was applied for this
        request.

        Parameters
        ----------
        request : aiohttp.web.Request
            The current request.
        response : aiohttp.web.Response
            The current response. May be modified in-place.
        context : object or None
            If present, the extra return value of ``authenticate``, providing
            any additional context needed to complete the authentication
            process.
        """
        pass


def unauthorized(kind):
    return web.HTTPUnauthorized(
        headers={"WWW-Authenticate": kind}, reason="Authentication required"
    )


class SimpleAuthenticator(Authenticator):
    """A simple authenticator that uses Basic Auth.

    This is highly insecure, use only for testing!!!
    """

    password = Unicode(
        None,
        allow_none=True,
        help="""
        If set, a global password that all users must provide.

        If unset (default), the password field is completely ignored.
        """,
        config=True,
    )

    async def authenticate(self, request):
        auth_header = request.headers.get("Authorization")
        if not auth_header:
            raise unauthorized("Basic")

        auth_type, auth_key = auth_header.split(" ", 1)
        if auth_type != "Basic":
            raise unauthorized("Basic")

        auth_key = base64.b64decode(auth_key).decode("ascii")
        user, password = auth_key.split(":", 1)

        if self.password and password != self.password:
            raise unauthorized("Basic")

        return User(user)


class KerberosAuthenticator(Authenticator):
    """An authenticator using kerberos"""

    service_name = Unicode(
        "HTTP",
        help="""The service's kerberos principal name.

        This is almost always "HTTP" (the default)""",
        config=True,
    )

    keytab = Unicode(
        "dask_gateway.keytab", help="The path to the keytab file", config=True
    )

    async def setup(self, app):
        os.environ["KRB5_KTNAME"] = self.keytab

    def raise_auth_error(self, err):
        self.log.error("Kerberos failure: %s", err)
        raise web.HTTPInternalServerError(reason="Error during kerberos authentication")

    async def authenticate(self, request):
        import kerberos

        auth_header = request.headers.get("Authorization")
        if not auth_header:
            raise unauthorized("Negotiate")

        auth_type, auth_key = auth_header.split(" ", 1)
        if auth_type != "Negotiate":
            raise unauthorized("Negotiate")

        gss_context = None
        try:
            # Initialize kerberos context
            rc, gss_context = kerberos.authGSSServerInit(self.service_name)

            # NOTE: Per the pykerberos documentation, the return code should be
            # checked after each step. However, after reading the pykerberos
            # code no method used here will ever return anything but
            # AUTH_GSS_COMPLETE (all other cases will raise an exception).  We
            # keep these checks in just in case pykerberos changes its behavior
            # to match its docs, but they likely never will trigger.

            if rc != kerberos.AUTH_GSS_COMPLETE:
                self.raise_auth_error("GSS server init failed, return code = %r" % rc)

            # Challenge step
            rc = kerberos.authGSSServerStep(gss_context, auth_key)
            if rc != kerberos.AUTH_GSS_COMPLETE:
                self.raise_auth_error("GSS server step failed, return code = %r" % rc)
            gss_key = kerberos.authGSSServerResponse(gss_context)

            # Retrieve user name
            fulluser = kerberos.authGSSServerUserName(gss_context)
            user = fulluser.split("@", 1)[0]
        except kerberos.GSSError as err:
            self.raise_auth_error(err)
        finally:
            if gss_context is not None:
                kerberos.authGSSServerClean(gss_context)

        return User(user), gss_key

    async def pre_response(self, request, response, context):
        response.headers["WWW-Authenticate"] = "Negotiate %s" % context


class JupyterHubAuthenticator(Authenticator):
    """An authenticator that uses JupyterHub to perform authentication"""

    jupyterhub_api_token = Unicode(
        help="""
        Dask Gateway's JupyterHub API Token, used for authenticating the
        gateway's API requests to JupyterHub.

        By default this is determined from the ``JUPYTERHUB_API_TOKEN``
        environment variable.
        """,
        config=True,
    )

    @default("jupyterhub_api_token")
    def _default_jupyterhub_api_token(self):
        out = os.environ.get("JUPYTERHUB_API_TOKEN")
        if not out:
            raise ValueError("JUPYTERHUB_API_TOKEN must be set")
        return out

    jupyterhub_api_url = Unicode(
        help="""
        The API URL for the JupyterHub server.

        By default this is determined from the ``JUPYTERHUB_API_URL``
        environment variable.
        """,
        config=True,
    )

    @default("jupyterhub_api_url")
    def _default_jupyterhub_api_url(self):
        out = os.environ.get("JUPYTERHUB_API_URL")
        if not out:
            raise ValueError("JUPYTERHUB_API_URL must be set")
        return out

    tls_key = Unicode(
        "",
        help="""
        Path to TLS key file for making API requests to JupyterHub.

        When setting this, you should also set tls_cert.
        """,
        config=True,
    )

    tls_cert = Unicode(
        "",
        help="""
        Path to TLS certficate file for making API requests to JupyterHub.

        When setting this, you should also set tls_cert.
        """,
        config=True,
    )

    tls_ca = Unicode(
        "",
        help="""
        Path to TLS CA file for verifying API requests to JupyterHub.

        When setting this, you should also set tls_key and tls_cert.
        """,
        config=True,
    )

    async def setup(self, app):
        self.session = aiohttp.ClientSession()
        if self.tls_cert and self.tls_key:
            import ssl

            ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            ssl_context.load_cert_chain(self.tls_cert, keyfile=self.tls_key)
            if self.tls_ca:
                ssl_context.load_verify_locations(self.tls_ca)
        else:
            ssl_context = None
        self.ssl_context = ssl_context

    async def cleanup(self):
        if hasattr(self, "session"):
            await self.session.close()

    def get_token(self, request):
        auth_header = request.headers.get("Authorization")
        if auth_header:
            auth_type, auth_key = auth_header.split(" ", 1)
            if auth_type == "jupyterhub":
                return auth_key
        return None

    async def authenticate(self, request):
        token = self.get_token(request)
        if token is None:
            raise unauthorized("jupyterhub")

        url = "{}/authorizations/token/{}".format(
            self.jupyterhub_api_url,
            quote(token, safe=""),
        )

        kwargs = {
            "headers": {"Authorization": "token %s" % self.jupyterhub_api_token},
            "ssl": self.ssl_context,
        }

        resp = await self.session.get(url, **kwargs)

        if resp.status < 400:
            data = await resp.json()
            # "groups" attribute doesn't exists in case of a service
            return User(
                data["name"],
                groups=data.get("groups", []),
                admin=data.get("admin", False),
            )
        elif resp.status == 404:
            self.log.debug("Token for non-existent user requested")
            raise unauthorized("jupyterhub")
        else:
            if resp.status == 403:
                err = web.HTTPInternalServerError(
                    reason="Permission failure verifying user's JupyterHub API token"
                )
            elif resp.status >= 500:
                err = web.HTTPBadGateway(
                    reason="Upstream failure verifying user's JupyterHub API token"
                )
            else:
                err = web.HTTPInternalServerError(
                    reason="Failure verifying user's JupyterHub API token"
                )

            self.log.error(
                "%s - code: %s, reason: %s", err.reason, resp.status, resp.reason
            )
            raise err



import asyncio
import json
import os
import subprocess
import uuid

from aiohttp import web
from traitlets import (
    Bool,
    CaselessStrEnum,
    Float,
    Integer,
    List,
    Unicode,
    default,
    validate,
)
from traitlets.config import Application, LoggingConfigurable, catch_config_error

from .. import __version__ as VERSION
from ..utils import CancelGroup, Flag, TaskPool, normalize_address

__all__ = ("Proxy", "ProxyApp")


_PROXY_EXE = os.path.join(
    os.path.abspath(os.path.dirname(os.path.relpath(__file__))), "dask-gateway-proxy"
)


class Proxy(LoggingConfigurable):
    """The dask-gateway-server proxy"""

    log_level = CaselessStrEnum(
        ["error", "warn", "info", "debug"],
        default_value="warn",
        help="The proxy log-level.",
        config=True,
    )

    address = Unicode(
        ":8000",
        help="""
        The address the HTTP/HTTPS proxy should *listen* at.

        Should be of the form ``{hostname}:{port}``.

        Where:

        - ``hostname`` sets the hostname to *listen* at. Set to ``""`` or
          ``"0.0.0.0"`` to listen on all interfaces.
        - ``port`` sets the port to *listen* at.
        """,
        config=True,
    )

    tcp_address = Unicode(
        help="""
        The address the TCP (scheduler) proxy should *listen* at.

        Should be of the form ``{hostname}:{port}``

        Where:

        - ``hostname`` sets the hostname to *listen* at. Set to ``""`` or
          ``"0.0.0.0"`` to listen on all interfaces.
        - ``port`` sets the port to *listen* at.

        If not specified, will default to `address`.
        """,
        config=True,
    )

    @default("tcp_address")
    def _default_tcp_address(self):
        return self.address

    @validate("address", "tcp_address")
    def _validate_addresses(self, proposal):
        return normalize_address(proposal.value)

    prefix = Unicode(
        "",
        help="""
        The path prefix the HTTP/HTTPS proxy should serve under.

        This prefix will be prepended to all routes registered with the proxy.
        """,
        config=True,
    )

    @validate("prefix")
    def _validate_prefix(self, proposal):
        prefix = proposal.value

        if prefix == "":
            return prefix

        if not prefix.startswith("/"):
            raise ValueError("invalid prefix, must start with ``/``")

        return prefix.rstrip("/")

    gateway_url = Unicode(
        help="The base URL the proxy should use for connecting to the gateway server",
        config=True,
    )

    @default("gateway_url")
    def _gateway_url_default(self):
        address = normalize_address(self.gateway_address, resolve_host=True)
        return f"http://{address}"

    api_token = Unicode(
        help="""
        The Proxy api token

        A 32 byte hex-encoded random string. Commonly created with the
        ``openssl`` CLI:

        .. code-block:: shell

            $ openssl rand -hex 32

        Loaded from the DASK_GATEWAY_PROXY_TOKEN env variable by default.
        """,
        config=True,
    )

    @default("api_token")
    def _api_token_default(self):
        token = os.environ.get("DASK_GATEWAY_PROXY_TOKEN", "")
        if not token:
            self.log.info("Generating new api token for proxy")
            token = uuid.uuid4().hex
        return token

    max_events = Integer(
        100,
        help="""
        The maximum number of events (proxy changes) to retain.

        A proxy server that lags by more than this number will have to do a
        full refesh.
        """,
        config=True,
    )

    tls_key = Unicode(
        "",
        help="""Path to TLS key file for the public url of the HTTP proxy.

        When setting this, you should also set tls_cert.
        """,
        config=True,
    )

    tls_cert = Unicode(
        "",
        help="""Path to TLS certificate file for the public url of the HTTP proxy.

        When setting this, you should also set tls_key.
        """,
        config=True,
    )

    externally_managed = Bool(
        False,
        help="""
        Whether the proxy process is externally managed.

        If False (default), the proxy process will be started and stopped by
        the gateway process. Set to True if the proxy will be started via some
        external manager (e.g. supervisord).
        """,
        config=True,
    )

    proxy_status_period = Float(
        30,
        help="""
        Time (in seconds) between proxy status checks.

        Only applies when ``externally_managed`` is False.
        """,
        config=True,
    )

    # forwarded from parent
    gateway_address = Unicode()

    @default("gateway_address")
    def _gateway_address_default(self):
        return self.parent.gateway_address

    def get_start_command(self, is_child_process=True):
        out = [
            _PROXY_EXE,
            "-address",
            self.address,
            "-tcp-address",
            self.tcp_address,
            "-api-url",
            self.gateway_url + "/api/v1/routes",
            "-log-level",
            self.log_level,
        ]
        if is_child_process:
            out.append("-is-child-process")
        if bool(self.tls_cert) != bool(self.tls_key):
            raise ValueError("Must set both tls_cert and tls_key")
        if self.tls_cert:
            out.extend(["-tls-cert", self.tls_cert, "-tls-key", self.tls_key])
        return out

    def get_start_env(self):
        env = os.environ.copy()
        env["DASK_GATEWAY_PROXY_TOKEN"] = self.api_token
        return env

    def start_proxy_process(self):
        command = self.get_start_command()
        env = self.get_start_env()
        self.log.info("Starting the Dask gateway proxy...")
        proc = subprocess.Popen(
            command,
            env=env,
            stdin=subprocess.PIPE,
            stdout=None,
            stderr=None,
            start_new_session=True,
        )
        self.proxy_process = proc
        self.log.info("Dask gateway proxy started")
        self.log.info(
            "- %s routes listening at %s://%s",
            "HTTPS" if self.tls_cert else "HTTP",
            "https" if self.tls_cert else "http",
            self.address,
        )
        self.log.info("- Scheduler routes listening at gateway://%s", self.tcp_address)

    async def monitor_proxy_process(self):
        backoff = default_backoff = 0.5
        while True:
            retcode = self.proxy_process.poll()
            if retcode is not None:
                self.log.warning(
                    "Proxy process exited unexpectedly with return code %d, restarting",
                    retcode,
                )
                try:
                    self.start_proxy_process()
                except Exception:
                    self.log.error(
                        "Failed starting the proxy process, retrying in %.1f seconds...",
                        backoff,
                        exc_info=True,
                    )
                    await asyncio.sleep(backoff)
                    backoff = min(backoff * 2, 15)
                    continue

            backoff = default_backoff
            await asyncio.sleep(self.proxy_status_period)

    async def setup(self, app):
        """Start the proxy."""
        self.cg = CancelGroup()
        self.task_pool = TaskPool()
        # Exposed for testing
        self._proxy_contacted = Flag()
        if not self.externally_managed:
            self.start_proxy_process()
            self.task_pool.spawn(self.monitor_proxy_process())

        # Internal state
        self.routes = {}
        self.offset = 0
        self.events = []
        self._watchers = set()
        self._next_id = 1

        app.on_shutdown.append(self._on_shutdown)

        app.add_routes([web.get("/api/v1/routes", self.routes_handler)])

        # Proxy through the gateway application
        await self.add_route(kind="PATH", path="/", target=self.gateway_url)

    async def _on_shutdown(self, app):
        await self.task_pool.close()
        await self.cg.cancel()

    async def cleanup(self):
        """Stop the proxy."""
        await self.task_pool.close()
        if hasattr(self, "proxy_process"):
            self.log.info("Stopping the Dask gateway proxy")
            self.proxy_process.terminate()

    def _get_id(self):
        o = self._next_id
        self._next_id += 1
        return o

    def _append_event(self, kind, route):
        event = {"id": self._get_id(), "type": kind, "route": route}

        if len(self.events) >= self.max_events:
            n_dropped = self.max_events // 2
            self.events = self.events[n_dropped:]
            self.offset += n_dropped

        self.events.append(event)

        for q in self._watchers:
            q.put_nowait([event])

    async def add_route(self, kind=None, sni=None, path=None, target=None):
        if target is None:
            raise ValueError("must specify `target`")

        if kind == "SNI":
            if sni is None:
                raise ValueError("must specify `sni`")
            route = {"kind": kind, "sni": sni, "target": target}
            self.routes[(kind, sni)] = route
        elif kind == "PATH":
            if path is None:
                raise ValueError("must specify `path`")
            path = self.prefix + path
            route = {"kind": kind, "path": path, "target": target}
            self.routes[(kind, path)] = route
        else:
            raise ValueError(f"Unknown route kind {kind}")
        self._append_event("PUT", route)

    async def remove_route(self, kind=None, sni=None, path=None):
        if kind == "SNI":
            if sni is None:
                raise ValueError("must specify `sni`")
            route = self.routes.pop((kind, sni), None)
            if route is not None:
                route = {"kind": kind, "sni": sni}
        elif kind == "PATH":
            if path is None:
                raise ValueError("must specify `path`")
            path = self.prefix + path
            route = self.routes.pop((kind, path), None)
            if route is not None:
                route = {"kind": kind, "path": path}
        else:
            raise ValueError(f"Unknown route kind {kind}")

        if route is not None:
            self._append_event("DELETE", route)

    def _events_after(self, last_id=0):
        index = last_id - self.offset
        if index >= 0:
            events = self.events[index:]
            return events
        else:
            # We've since dropped these events, client needs to reset
            return None

    async def routes_handler(self, request):
        self._proxy_contacted.set()
        # Authenticate the api request
        auth_header = request.headers.get("Authorization")
        if not auth_header:
            raise web.HTTPUnauthorized(headers={"WWW-Authenticate": "token"})

        auth_type, auth_token = auth_header.split(" ", 1)
        if auth_type != "token" or auth_token != self.api_token:
            raise web.HTTPUnauthorized(headers={"WWW-Authenticate": "token"})

        watch = request.query.get("watch", False)

        if watch:
            last_id = request.query.get("last_id", 0)
            try:
                last_id = int(last_id)
            except Exception:
                raise web.HTTPUnprocessableEntity(
                    reason="invalid `last_id`: %s" % last_id
                )

            events = self._events_after(last_id)
            if events is None:
                raise web.HTTPGone(reason="`last_id` is too old, full refresh required")

            queue = asyncio.Queue()
            if events:
                queue.put_nowait(events)
            self._watchers.add(queue)
            try:
                response = web.StreamResponse(
                    headers={"Content-Type": "application/json"}
                )
                await response.prepare(request)

                async with self.cg.cancellable():
                    while True:
                        events = await queue.get()
                        msg = {"events": events}
                        await response.write(json.dumps(msg).encode() + b"\n")
            finally:
                self._watchers.discard(queue)

            await response.write_eof()
            return response

        else:
            msg = {"routes": list(self.routes.values()), "id": self._next_id - 1}
            return web.json_response(msg)


class ProxyApp(Application):
    """Start a proxy application"""

    name = "dask-gateway-server proxy"
    description = "Start the proxy"
    examples = """

        dask-gateway-server proxy -f config.py
    """
    version = VERSION

    config_file = Unicode(
        "dask_gateway_config.py", help="The config file to load", config=True
    )

    aliases = {
        "f": "ProxyApp.config_file",
        "config": "ProxyApp.config_file",
        "log-level": "Proxy.log_level",
    }

    classes = List([Proxy])

    @catch_config_error
    def initialize(self, argv=None):
        super().initialize(argv)
        self.parent.load_config_file(self.config_file)
        self.parent.config.merge(self.config)

    def start(self):
        proxy = Proxy(
            parent=self.parent, log=self.parent.log, gateway_address=self.parent.address
        )
        command = proxy.get_start_command(is_child_process=False)
        env = proxy.get_start_env()
        exe = command[0]
        args = command[1:]
        os.execle(exe, "dask-gateway-proxy", *args, env)



import asyncio
import functools
import json
from datetime import datetime, timezone

from kubernetes_asyncio import client
from traitlets import Any, Dict, Int, Unicode
from traitlets.config import LoggingConfigurable

from ...traitlets import Callable
from ...utils import cancel_task

# Monkeypatch kubernetes_asyncio to cleanup resources better
client.rest.RESTClientObject.__del__ = lambda self: None


def parse_k8s_timestamp(ts):
    t = datetime.fromisoformat(ts[:-1])
    return int(t.replace(tzinfo=timezone.utc).timestamp() * 1000)


def k8s_timestamp():
    t = datetime.now(timezone.utc)
    return t.strftime("%Y-%m-%dT%H:%M:%SZ")


class RateLimitedClient:
    """A rate limited wrapper around a kubernetes client"""

    def __init__(self, client, rate_limiter):
        self.rate_limiter = rate_limiter
        self.client = client
        self._cache = {}

    @property
    def api_client(self):
        return self.client.api_client

    def __getattr__(self, name):
        if not name.startswith("_") and hasattr(self.client, name):
            if name not in self._cache:
                method = getattr(self.client, name)

                @functools.wraps(method)
                async def func(*args, **kwargs):
                    await self.rate_limiter.acquire()
                    return await method(*args, **kwargs)

                self._cache[name] = func
            return self._cache[name]
        raise AttributeError(name)


async def watch(method, *args, **kwargs):
    """Handle a single watch call ona resource"""
    kwargs["watch"] = True
    kwargs["_preload_content"] = False

    resp = await method(*args, **kwargs)
    async with resp:
        while True:
            line = await resp.content.readline()
            line = line.decode("utf8")
            if not line:
                break
            yield json.loads(line)


class Informer(LoggingConfigurable):
    name = Unicode()
    client = Any(None)
    method = Unicode()
    method_kwargs = Dict()
    timeout_seconds = Int(10)
    on_update = Callable()
    on_delete = Callable()

    def get(self, name, default=None):
        return self.cache.get(name, default)

    def get_key(self, obj):
        namespace = obj["metadata"]["namespace"]
        name = obj["metadata"]["name"]
        return f"{namespace}.{name}"

    def handle_initial(self, objs):
        obj_map = {self.get_key(i): i for i in objs}
        if self.cache:
            deleted = set(self.cache).difference(obj_map)
        else:
            deleted = ()
        self.log.debug(
            "Relisted %s informer - %d deletes, %d updates",
            self.name,
            len(deleted),
            len(objs),
        )
        for key in deleted:
            obj = self.cache.pop(key)
            try:
                self.on_delete(obj)
            except Exception:
                self.log.warning("Error in %s informer", self.name, exc_info=True)
        for key, obj in obj_map.items():
            old = self.cache.get(key)
            self.cache[key] = obj
            try:
                self.on_update(obj, old=old)
            except Exception:
                self.log.warning("Error in %s informer", self.name, exc_info=True)

    def handle(self, obj, event_type="ADDED"):
        key = self.get_key(obj)
        self.log.debug("Event - %s %s %s", event_type, self.name, key)

        old = self.cache.get(key)
        if event_type in {"ADDED", "MODIFIED"}:
            self.cache[key] = obj
            try:
                self.on_update(obj, old=old)
            except Exception:
                self.log.warning("Error in %s informer", self.name, exc_info=True)
        elif event_type == "DELETED":
            self.cache.pop(key, None)
            try:
                self.on_delete(obj)
            except Exception:
                self.log.warning("Error in %s informer", self.name, exc_info=True)

    async def start(self):
        if not hasattr(self, "_task"):
            self.cache = {}
            self.initialized = asyncio.Event()
            self._task = asyncio.ensure_future(self.run())
        await self.initialized.wait()

    async def stop(self):
        if hasattr(self, "_task"):
            await cancel_task(self._task)
            del self._task

    async def run(self):
        self.log.debug("Starting %s informer...", self.name)

        method = getattr(self.client, self.method)

        while True:
            try:
                initial = await method(**self.method_kwargs)
            except asyncio.CancelledError:
                raise
            except Exception as exc:
                self.log.error(
                    "Error in %s informer, retrying...", self.name, exc_info=exc
                )
                continue

            initial = self.client.api_client.sanitize_for_serialization(initial)
            self.handle_initial(initial["items"])
            self.initialized.set()
            res_ver = initial["metadata"]["resourceVersion"]

            try:
                while True:
                    relist = False
                    kwargs = {
                        "resource_version": res_ver,
                        "timeout_seconds": self.timeout_seconds,
                    }
                    kwargs.update(self.method_kwargs)
                    async for ev in watch(method, **kwargs):
                        typ = ev["type"]
                        obj = ev["object"]
                        if typ.lower() == "error":
                            code = obj.get("code", 0)
                            if code == 410:
                                self.log.debug(
                                    "Too old resourceVersion in %s informer, relisting",
                                    self.name,
                                )
                            else:
                                self.log.warning(
                                    (
                                        "Unexpected error in %s informer, relisting: "
                                        "message=%r, code=%d"
                                    ),
                                    obj.get("message"),
                                    code,
                                )
                            relist = True
                            break
                        try:
                            res_ver = ev["object"]["metadata"]["resourceVersion"]
                        except KeyError:
                            pass
                        self.handle(obj, typ)
                    if relist:
                        break
            except asyncio.CancelledError:
                raise
            except Exception as exc:
                self.log.error(
                    "Error in %s informer, retrying...", self.name, exc_info=exc
                )
        self.log.debug("%s informer stopped", self.name)


def merge_json_objects(a, b):
    """Merge two JSON objects recursively.

    - If a dict, keys are merged, preferring ``b``'s values
    - If a list, values from ``b`` are appended to ``a``

    Copying is minimized. No input collection will be mutated, but a deep copy
    is not performed.

    Parameters
    ----------
    a, b : dict
        JSON objects to be merged.

    Returns
    -------
    merged : dict
    """
    if b:
        # Use a shallow copy here to avoid needlessly copying
        a = a.copy()
        for key, b_val in b.items():
            if key in a:
                a_val = a[key]
                if isinstance(a_val, dict) and isinstance(b_val, dict):
                    a[key] = merge_json_objects(a_val, b_val)
                elif isinstance(a_val, list) and isinstance(b_val, list):
                    a[key] = a_val + b_val
                else:
                    a[key] = b_val
            else:
                a[key] = b_val
    return a



import asyncio
import os
import uuid
from base64 import b64decode
from collections import defaultdict

from kubernetes_asyncio import client, config
from kubernetes_asyncio.client.rest import ApiException
from traitlets import Dict, Float, Unicode, default
from traitlets.config import LoggingConfigurable

from ... import __version__ as VERSION
from ... import models
from ...traitlets import MemoryLimit, Type
from ...utils import Flag
from ...workqueue import WorkQueue, WorkQueueClosed
from ..base import Backend, ClusterConfig
from .utils import Informer, parse_k8s_timestamp

__all__ = ("KubeClusterConfig", "KubeBackend", "KubeBackendAndControllerMixin")


class KubeClusterConfig(ClusterConfig):
    """Configuration for a single Dask cluster running on kubernetes"""

    namespace = Unicode(
        "default",
        help="""
        Kubernetes namespace to launch pods in.

        If running inside a kubernetes cluster with service accounts enabled,
        defaults to the current namespace. If not, defaults to `default`
        """,
        config=True,
    )

    @default("namespace")
    def _default_namespace(self):
        ns_path = "/var/run/secrets/kubernetes.io/serviceaccount/namespace"
        if os.path.exists(ns_path):
            with open(ns_path) as f:
                return f.read().strip()
        return "default"

    image = Unicode(
        "daskgateway/dask-gateway:latest",
        help="Docker image to use for running user's containers.",
        config=True,
    )

    image_pull_policy = Unicode(
        "IfNotPresent",
        help="The image pull policy of the docker image specified in ``image``",
        config=True,
    )

    # Kubernetes is a bit different in types/granularity of resource requests.
    # We redefine these common fields here to support that.
    worker_cores = Float(
        1,
        min=0,
        help="""
        Number of cpu-cores available for a dask worker.
        """,
        config=True,
    )

    worker_cores_limit = Float(
        min=0,
        help="""
        Maximum number of cpu-cores available for a dask worker.

        Defaults to ``worker_cores``.
        """,
        config=True,
    )

    @default("worker_cores_limit")
    def _default_worker_cores_limit(self):
        return self.worker_cores

    worker_memory_limit = MemoryLimit(
        help="""
        Maximum number of bytes available for a dask worker. Allows the
        following suffixes:

        - K -> Kibibytes
        - M -> Mebibytes
        - G -> Gibibytes
        - T -> Tebibytes

        Defaults to ``worker_memory``.
        """,
        config=True,
    )

    @default("worker_memory_limit")
    def _default_worker_memory_limit(self):
        return self.worker_memory

    scheduler_cores = Float(
        1,
        min=0,
        help="""
        Number of cpu-cores available for a dask scheduler.
        """,
        config=True,
    )

    scheduler_cores_limit = Float(
        min=0,
        help="""
        Maximum number of cpu-cores available for a dask scheduler.

        Defaults to ``scheduler_cores``.
        """,
        config=True,
    )

    @default("scheduler_cores_limit")
    def _default_scheduler_cores_limit(self):
        return self.scheduler_cores

    scheduler_memory_limit = MemoryLimit(
        help="""
        Maximum number of bytes available for a dask scheduler. Allows the
        following suffixes:

        - K -> Kibibytes
        - M -> Mebibytes
        - G -> Gibibytes
        - T -> Tebibytes

        Defaults to ``scheduler_memory``.
        """,
        config=True,
    )

    @default("scheduler_memory_limit")
    def _default_scheduler_memory_limit(self):
        return self.scheduler_memory

    worker_extra_container_config = Dict(
        help="""
        Any extra configuration for the worker container.

        This dict will be deep merged with the worker container (a
        ``V1Container`` object) before submission. Keys should match those in
        the `kubernetes spec
        <https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#container-v1-core>`__,
        and should be camelCase.

        For example, here we add environment variables from a secret to the
        worker container:

        .. code::

            c.KubeClusterConfig.worker_extra_container_config = {
                "envFrom": [
                    {"secretRef": {"name": "my-env-secret"}}
                ]
            }
        """,
        config=True,
    )

    scheduler_extra_container_config = Dict(
        help="""
        Any extra configuration for the scheduler container.

        This dict will be deep merged with the scheduler container (a
        ``V1Container`` object) before submission. Keys should match those in
        the `kubernetes spec
        <https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#container-v1-core>`__,
        and should be camelCase.

        See ``worker_extra_container_config`` for more information.
        """,
        config=True,
    )

    worker_extra_pod_config = Dict(
        help="""
        Any extra configuration for the worker pods.

        This dict will be deep merged with the worker pod spec (a ``V1PodSpec``
        object) before submission. Keys should match those in the `kubernetes
        spec
        <https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#podspec-v1-core>`__,
        and should be camelCase.

        For example, here we add a toleration to worker pods.

        .. code::

            c.KubeClusterConfig.worker_extra_pod_config = {
                "tolerations": [
                    {
                        "key": "key",
                        "operator": "Equal",
                        "value": "value",
                        "effect": "NoSchedule",
                    }
                ]
            }
        """,
        config=True,
    )

    scheduler_extra_pod_config = Dict(
        help="""
        Any extra configuration for the scheduler pods.

        This dict will be deep merged with the scheduler pod spec (a
        ``V1PodSpec`` object) before submission. Keys should match those in the
        `kubernetes spec
        <https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#podspec-v1-core>`__,
        and should be camelCase.

        See ``worker_extra_pod_config`` for more information.
        """,
        config=True,
    )

    worker_extra_pod_annotations = Dict(
        help="""
        Any extra annotations to be applied to a user's worker pods.

        These annotations can be set using with cluster options (See
        :ref:`exposing-cluster-options`) to allow for injecting user-specific
        information, e.g. adding an annotation based on a user's group or
        username.

        This dict will be merged with ``common_annotations`` before being
        applied to user pods.
        """,
        config=True,
    )

    scheduler_extra_pod_annotations = Dict(
        help="""
        Any extra annotations to be applied to a user's scheduler pods.

        These annotations can be set using with cluster options (See
        :ref:`exposing-cluster-options`) to allow for injecting user-specific
        information, e.g. adding an annotation based on a user's group or
        username.

        This dict will be merged with ``common_annotations`` before being
        applied to user pods.
        """,
        config=True,
    )

    worker_extra_pod_labels = Dict(
        help="""
        Any extra labels to be applied to a user's worker pods.

        These labels can be set using with cluster options (See
        :ref:`exposing-cluster-options`) to allow for injecting user-specific
        information, e.g. adding a label based on a user's group or username.

        This dict will be merged with ``common_labels`` before being
        applied to user pods.
        """,
        config=True,
    )

    scheduler_extra_pod_labels = Dict(
        help="""
        Any extra labels to be applied to a user's scheduler pods.

        These labels can be set using with cluster options (See
        :ref:`exposing-cluster-options`) to allow for injecting user-specific
        information, e.g. adding a label based on a user's group or username.

        This dict will be merged with ``common_labels`` before being
        applied to user pods.
        """,
        config=True,
    )


class KubeBackendAndControllerMixin(LoggingConfigurable):
    """Shared config between the backend and controller"""

    gateway_instance = Unicode(
        help="""
        A unique ID for this instance of dask-gateway.

        The controller must also be configured with the same ID.
        """,
        config=True,
    )

    @default("gateway_instance")
    def _default_instance(self):
        instance = os.environ.get("DASK_GATEWAY_INSTANCE")
        if not instance:
            raise ValueError("Must specify `c.KubeBackend.instance`")
        return instance

    crd_version = Unicode(
        "v1alpha1", help="The version for the DaskCluster CRD", config=True
    )

    common_labels = Dict(
        {
            "app.kubernetes.io/name": "dask-gateway",
            "app.kubernetes.io/version": VERSION.replace("+", "_"),
        },
        help="Kubernetes labels to apply to all objects created by the gateway",
        config=True,
    )

    common_annotations = Dict(
        help="Kubernetes annotations to apply to all objects created by the gateway",
        config=True,
    )

    label_selector = Unicode(
        help="""
        The label selector to use when watching objects managed by the gateway.
        """,
        config=True,
    )

    @default("label_selector")
    def _default_label_selector(self):
        return f"gateway.dask.org/instance={self.gateway_instance}"


class KubeBackend(KubeBackendAndControllerMixin, Backend):
    """A dask-gateway backend for running on Kubernetes"""

    cluster_config_class = Type(
        "dask_gateway_server.backends.kubernetes.backend.KubeClusterConfig",
        klass="dask_gateway_server.backends.base.ClusterConfig",
        help="The cluster config class to use",
        config=True,
    )

    async def setup(self, app):
        await super().setup(app)
        try:
            # Not a coroutine for some reason
            config.load_incluster_config()
        except config.ConfigException:
            await config.load_kube_config()

        self.api_client = client.ApiClient()
        self.core_client = client.CoreV1Api(api_client=self.api_client)
        self.custom_client = client.CustomObjectsApi(api_client=self.api_client)

        self.cluster_waiters = defaultdict(Flag)
        self.clusters = {}
        self.username_to_clusters = defaultdict(dict)
        self.queue = WorkQueue()
        self.informer = Informer(
            parent=self,
            name="cluster",
            client=self.custom_client,
            method="list_cluster_custom_object",
            method_kwargs=dict(
                group="gateway.dask.org",
                version=self.crd_version,
                plural="daskclusters",
                label_selector=self.label_selector,
            ),
            on_update=self.on_cluster_event,
            on_delete=self.on_cluster_event,
        )
        await self.informer.start()
        self.sync_task = asyncio.ensure_future(self.sync_clusters_loop())

    async def cleanup(self):
        if hasattr(self, "informer"):
            await self.informer.stop()
        if hasattr(self, "sync_task"):
            self.queue.close()
            try:
                await self.sync_task
            except Exception:
                pass
        if hasattr(self, "api_client"):
            await self.api_client.rest_client.pool_manager.close()
        await super().cleanup()

    async def start_cluster(self, user, cluster_options):
        options, config = await self.process_cluster_options(user, cluster_options)

        obj = self.make_cluster_object(user.name, options, config)

        name = obj["metadata"]["name"]
        cluster_name = f"{config.namespace}.{name}"

        self.log.info("Creating cluster %s for user %s", cluster_name, user.name)

        await self.custom_client.create_namespaced_custom_object(
            "gateway.dask.org", self.crd_version, config.namespace, "daskclusters", obj
        )
        return cluster_name

    async def stop_cluster(self, cluster_name, failed=False):
        cluster = self.clusters.get(cluster_name)
        if cluster is not None and cluster.status >= models.ClusterStatus.STOPPED:
            # We know for sure the cluster was already stopped, do nothing
            return

        self.log.info("Stopping cluster %s", cluster_name)

        namespace, name = cluster_name.split(".")
        await self.custom_client.patch_namespaced_custom_object(
            "gateway.dask.org",
            self.crd_version,
            namespace,
            "daskclusters",
            name,
            [{"op": "add", "path": "/spec/active", "value": False}],
        )

    async def get_cluster(self, cluster_name, wait=False):
        if wait:
            try:
                waiter = self.cluster_waiters[cluster_name]
                await asyncio.wait_for(waiter, 20)
            except asyncio.TimeoutError:
                pass
        return self.clusters.get(cluster_name)

    async def list_clusters(self, username=None, statuses=None):
        if statuses is None:
            select = lambda x: x.status <= models.ClusterStatus.RUNNING
        else:
            statuses = set(statuses)
            select = lambda x: x.status in statuses
        if username is None:
            return [cluster for cluster in self.clusters.values() if select(cluster)]
        else:
            clusters = self.username_to_clusters.get(username)
            if clusters is None:
                return []
            return [cluster for cluster in clusters.values() if select(cluster)]

    async def on_cluster_heartbeat(self, cluster_name, msg):
        count = msg["count"]
        namespace, name = cluster_name.split(".")

        self.log.info(
            "Cluster %s heartbeat [count: %d, n_active: %d, n_closing: %d, n_closed: %d]",
            cluster_name,
            count,
            len(msg["active_workers"]),
            len(msg["closing_workers"]),
            len(msg["closed_workers"]),
        )

        cluster = self.clusters.get(cluster_name)
        if cluster is None:
            return
        max_workers = cluster.config.get("cluster_max_workers")
        if max_workers is not None and count > max_workers:
            # This shouldn't happen under normal operation, but could if the
            # user does something malicious (or there's a bug).
            self.log.info(
                "Cluster %s heartbeat requested %d workers, exceeding limit of %s.",
                cluster_name,
                count,
                max_workers,
            )
            count = max_workers

        try:
            await self.custom_client.patch_namespaced_custom_object(
                "gateway.dask.org",
                self.crd_version,
                namespace,
                "daskclusters",
                name,
                [{"op": "add", "path": "/spec/replicas", "value": count}],
            )
        except ApiException as exc:
            if exc.status != 404:
                raise

    def get_labels(self, cluster_name, component=None):
        labels = self.common_labels.copy()
        labels.update(
            {
                "gateway.dask.org/instance": self.gateway_instance,
                "gateway.dask.org/cluster": cluster_name,
            }
        )
        if component:
            labels["app.kubernetes.io/component"] = component
        return labels

    def make_cluster_object(self, username, options, config):
        cluster_name = uuid.uuid4().hex

        return {
            "apiVersion": f"gateway.dask.org/{self.crd_version}",
            "kind": "DaskCluster",
            "metadata": {
                "name": cluster_name,
                "labels": self.get_labels(cluster_name),
                "annotations": self.common_annotations,
            },
            "spec": {
                "username": username,
                "options": options,
                "config": config.to_dict(),
            },
        }

    def get_cluster_name(self, obj):
        namespace = obj["metadata"]["namespace"]
        name = obj["metadata"]["name"]
        return f"{namespace}.{name}"

    def on_cluster_event(self, cluster, old=None):
        key = self.get_cluster_name(cluster)
        self.queue.put(key)

    async def sync_clusters_loop(self):
        while True:
            try:
                name = await self.queue.get()
            except WorkQueueClosed:
                return
            try:
                await self.sync_cluster(name)
            except Exception:
                self.log.warning("Error while syncing cluster %s", name, exc_info=True)
                self.queue.put_backoff(name)
            else:
                self.queue.reset_backoff(name)
            finally:
                self.queue.task_done(name)

    async def sync_cluster(self, cluster_name):
        obj = self.informer.get(cluster_name)

        if obj is None:
            self.log.debug("Deleting %s from cache", cluster_name)
            cluster = self.clusters.pop(cluster_name, None)
            if cluster is not None:
                user = self.username_to_clusters.get(cluster.username)
                if user is not None:
                    user.pop(cluster.name, None)
                    if not user:
                        self.username_to_clusters.pop(cluster.username, None)
            waiter = self.cluster_waiters.pop(cluster_name, None)
            if waiter is not None:
                waiter.set()
        else:
            self.log.debug("Updating %s in cache", cluster_name)

            old = self.clusters.get(cluster_name)

            status = obj.get("status", {})
            phase = status.get("phase", "Pending")
            cluster_status = models.ClusterStatus.from_name(phase)

            service_name = status.get("service")
            if cluster_status == models.ClusterStatus.RUNNING and service_name:
                namespace = obj["metadata"]["namespace"]
                scheduler_address = f"tls://{service_name}.{namespace}:8786"
                dashboard_address = f"http://{service_name}.{namespace}:8787"
                api_address = f"http://{service_name}.{namespace}:8788"
            else:
                scheduler_address = dashboard_address = api_address = ""

            start_time = parse_k8s_timestamp(obj["metadata"]["creationTimestamp"])
            if "completionTime" in status:
                stop_time = parse_k8s_timestamp(status["completionTime"])
            else:
                stop_time = None

            cluster = models.Cluster(
                name=cluster_name,
                username=obj["spec"].get("username", ""),
                options=obj["spec"].get("options") or {},
                config=obj["spec"].get("config") or {},
                token="",
                scheduler_address=scheduler_address,
                dashboard_address=dashboard_address,
                api_address=api_address,
                status=cluster_status,
                start_time=start_time,
                stop_time=stop_time,
            )

            if old is not None and old.tls_cert:
                cluster.tls_cert = old.tls_cert
                cluster.tls_key = old.tls_key
                cluster.token = old.token
            elif cluster_status <= models.ClusterStatus.RUNNING:
                secret_name = status.get("credentials")
                if secret_name:
                    namespace = obj["metadata"]["namespace"]
                    secret = await self.core_client.read_namespaced_secret(
                        secret_name, namespace
                    )
                    cluster.tls_cert = b64decode(secret.data["dask.crt"])
                    cluster.tls_key = b64decode(secret.data["dask.pem"])
                    cluster.token = b64decode(secret.data["api-token"]).decode()

            self.clusters[cluster.name] = cluster
            self.username_to_clusters[cluster.username][cluster.name] = cluster
            if cluster_status >= models.ClusterStatus.RUNNING:
                self.cluster_waiters[cluster.name].set()



import enum

__all__ = ("User", "Cluster", "ClusterStatus")


class IntEnum(enum.IntEnum):
    @classmethod
    def from_name(cls, name):
        """Create an enum value from a name"""
        try:
            return cls[name.upper()]
        except KeyError:
            pass
        raise ValueError(f"{name!r} is not a valid {cls.__name__}")


class ClusterStatus(IntEnum):
    """A cluster's status.

    Attributes
    ----------
    PENDING : ClusterStatus
        The cluster is pending start.
    RUNNING : ClusterStatus
        The cluster is running.
    STOPPING : ClusterStatus
        The cluster is stopping, but not fully stopped.
    STOPPED : ClusterStatus
        The cluster was shutdown by user or admin request.
    FAILED : ClusterStatus
        A failure occurred during any of the above states.
    """

    PENDING = 1
    RUNNING = 2
    STOPPING = 3
    STOPPED = 4
    FAILED = 5


class User:
    """A user record.

    Parameters
    ----------
    name : str
        The username
    groups : sequence, optional
        A set of groups the user belongs to. Default is no groups.
    admin : bool, optional
        Whether the user is an admin user. Default is False.
    """

    def __init__(self, name, groups=(), admin=False):
        self.name = name
        self.groups = set(groups)
        self.admin = bool(admin)

    def has_permissions(self, cluster):
        """Check if the user has permissions to view the cluster"""
        return self.name == cluster.username


class Cluster:
    """A cluster record.

    Parameters
    ----------
    name : str
        The cluster name. An unique string that can be used to identify the
        cluster in the gateway.
    username : str
        The username that started this cluster.
    token : str
        The API token associated with this cluster. Used to authenticate the
        cluster with the gateway.
    options : dict
        The normalized set of configuration options provided when starting this
        cluster. These values are user-facing, and don't necessarily correspond
        with the ``ClusterConfig`` options on the backend.
    config : dict
        The serialized version of ``ClusterConfig`` for this cluster.
    status : ClusterStatus
        The status of the cluster.
    scheduler_address : str
        The scheduler address. The empty string if the cluster is not running.
    dashboard_address : str
        The dashboard address. The empty string if the cluster is not running,
        or no dashboard is running on the cluster.
    api_address : str
        The cluster's api address. The empty string if the cluster is not running.
    tls_cert : bytes
        The TLS cert credentials associated with the cluster.
    tls_key : bytes
        The TLS key credentials associated with the cluster.
    start_time : int or None
        Start time in ms since the epoch, or None if not started.
    stop_time : int or None
        Stop time in ms since the epoch, or None if not stopped.
    """

    def __init__(
        self,
        name,
        username,
        token,
        options,
        config,
        status,
        scheduler_address="",
        dashboard_address="",
        api_address="",
        tls_cert=b"",
        tls_key=b"",
        start_time=None,
        stop_time=None,
    ):
        self.name = name
        self.username = username
        self.token = token
        self.options = options
        self.config = config
        self.status = status
        self.scheduler_address = scheduler_address
        self.dashboard_address = dashboard_address
        self.api_address = api_address
        self.tls_cert = tls_cert
        self.tls_key = tls_key
        self.start_time = start_time
        self.stop_time = stop_time

    def to_dict(self, full=True):
        if self.status == ClusterStatus.RUNNING and self.dashboard_address:
            dashboard = "/clusters/%s/status" % self.name
        else:
            dashboard = None
        out = {
            "name": self.name,
            "dashboard_route": dashboard,
            "status": self.status.name,
            "start_time": self.start_time,
            "stop_time": self.stop_time,
            "options": self.options,
        }
        if full:
            if self.status == ClusterStatus.RUNNING:
                tls_cert = self.tls_cert.decode()
                tls_key = self.tls_key.decode()
            else:
                tls_cert = tls_key = None
            out["tls_cert"] = tls_cert
            out["tls_key"] = tls_key
        return out



import math
import os
import shutil

from traitlets import Unicode, default

from ...traitlets import Type
from .base import JobQueueBackend, JobQueueClusterConfig

__all__ = ("SlurmBackend", "SlurmClusterConfig")


def slurm_format_memory(n):
    """Format memory in bytes for use with slurm."""
    if n >= 10 * (1024**3):
        return "%dG" % math.ceil(n / (1024**3))
    if n >= 10 * (1024**2):
        return "%dM" % math.ceil(n / (1024**2))
    if n >= 10 * 1024:
        return "%dK" % math.ceil(n / 1024)
    return "1K"


class SlurmClusterConfig(JobQueueClusterConfig):
    """Dask cluster configuration options when running on SLURM"""

    partition = Unicode("", help="The partition to submit jobs to.", config=True)

    qos = Unicode("", help="QOS string associated with each job.", config=True)

    account = Unicode("", help="Account string associated with each job.", config=True)


class SlurmBackend(JobQueueBackend):
    """A backend for deploying Dask on a Slurm cluster."""

    cluster_config_class = Type(
        "dask_gateway_server.backends.jobqueue.slurm.SlurmClusterConfig",
        klass="dask_gateway_server.backends.base.ClusterConfig",
        help="The cluster config class to use",
        config=True,
    )

    @default("submit_command")
    def _default_submit_command(self):
        return shutil.which("sbatch") or "sbatch"

    @default("cancel_command")
    def _default_cancel_command(self):
        return shutil.which("scancel") or "scancel"

    @default("status_command")
    def _default_status_command(self):
        return shutil.which("squeue") or "squeue"

    def get_submit_cmd_env_stdin(self, cluster, worker=None):
        cmd = [self.submit_command, "--parsable"]
        cmd.append("--job-name=dask-gateway")
        if cluster.config.partition:
            cmd.append("--partition=" + cluster.config.partition)
        if cluster.config.account:
            cmd.account("--account=" + cluster.config.account)
        if cluster.config.qos:
            cmd.extend("--qos=" + cluster.config.qos)

        if worker:
            cpus = cluster.config.worker_cores
            mem = slurm_format_memory(cluster.config.worker_memory)
            log_file = "dask-worker-%s.log" % worker.name
            script = "\n".join(
                [
                    "#!/bin/sh",
                    cluster.config.worker_setup,
                    " ".join(self.get_worker_command(cluster, worker.name)),
                ]
            )
            env = self.get_worker_env(cluster)
        else:
            cpus = cluster.config.scheduler_cores
            mem = slurm_format_memory(cluster.config.scheduler_memory)
            log_file = "dask-scheduler-%s.log" % cluster.name
            script = "\n".join(
                [
                    "#!/bin/sh",
                    cluster.config.scheduler_setup,
                    " ".join(self.get_scheduler_command(cluster)),
                ]
            )
            env = self.get_scheduler_env(cluster)

        staging_dir = self.get_staging_directory(cluster)

        cmd.extend(
            [
                "--chdir=" + staging_dir,
                "--output=" + os.path.join(staging_dir, log_file),
                "--cpus-per-task=%d" % cpus,
                "--mem=%s" % mem,
                "--export=%s" % (",".join(sorted(env))),
            ]
        )

        return cmd, env, script

    def get_stop_cmd_env(self, job_id):
        return [self.cancel_command, job_id], {}

    def get_status_cmd_env(self, job_ids):
        cmd = [self.status_command, "-h", "--job=%s" % ",".join(job_ids), "-o", "%i %t"]
        return cmd, {}

    def parse_job_states(self, stdout):
        states = {}
        for l in stdout.splitlines():
            job_id, state = l.split()
            states[job_id] = state in ("R", "CG", "PD", "CF")
        return states

    def parse_job_id(self, stdout):
        return stdout.strip()
